package main;

import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeVerifier;
import meerkat.crypto.mixnet.Mixer;
import meerkat.protobuf.BulletinBoardAPI;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;
import necessary.AsyncBulletinBoardClient;
import necessary.BulletinBoardClient;
import necessary.SignedBatch;
import qilin.util.Pair;


import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;


/**
 * Created by Tzlil on 12/17/2015.
 */
public class MainMixing {

    private Mixer mixer;
    private Mix2ZeroKnowledgeVerifier verifier;
    private int n, layers;
    private AsyncBulletinBoardClient asyncBulletinBoardClient;
    private final byte[] id;
    private final int mixerOrder;


    public MainMixing(Mixer mixer, int mixerOrder, Mix2ZeroKnowledgeVerifier verifier, int n
            , AsyncBulletinBoardClient asyncBulletinBoardClient, byte[] id) {
        this.mixer = mixer;
        this.mixerOrder = mixerOrder;
        this.verifier = verifier;
        this.n = n;
        this.layers = (int) (2 * Math.log(n) / Math.log(2)) - 1;  // layers = 2logn -1
        this.asyncBulletinBoardClient = asyncBulletinBoardClient;
        this.id = id;
    }

    public void main(List<Integer> prevBatchIds, int batchId, BulletinBoardClient.ClientCallback<?> callback) throws Exception {

        List<Crypto.RerandomizableEncryptedMessage> mixerInput;


        if (mixerOrder > 0) {
            List<BatchHandler> batchHandlers = new ArrayList<BatchHandler>(prevBatchIds.size());
            BatchHandler currentBatchHandler;
            for (Integer prevBatchId : prevBatchIds) {
                currentBatchHandler = new BatchHandler(n, layers);
                asyncBulletinBoardClient.readBatch(id, prevBatchId, currentBatchHandler);
                batchHandlers.add(currentBatchHandler);
            }

            boolean allDone = false;
            while (!allDone) {
                try {
                    Thread.sleep(30);
                } catch (InterruptedException e) {
                    // do nothing
                }
                // check all handlers done
                allDone = true;
                for (BatchHandler batchHandler : batchHandlers) {
                    allDone &= batchHandler.done;
                }
            }

            // assert all handlers succeeded
            for (BatchHandler batchHandler : batchHandlers) {
                if(batchHandler.e != null)
                    throw batchHandler.e;
            }

            BatchHandler lastBatchHandler = batchHandlers.get(batchHandlers.size() - 1);
            mixerInput = lastBatchHandler.getInputForMixer();

        } else {
            // ToDo : handle first mixer case
            mixerInput = null;
        }

        Pair<Mixing.ZeroKnowledgeProof[][], Crypto.RerandomizableEncryptedMessage[][]> mixerOutput
                = mixer.mix(mixerInput);
        updateBB(mixerOutput, batchId, callback);

    }


    private void updateBB(Pair<Mixing.ZeroKnowledgeProof[][], Crypto.RerandomizableEncryptedMessage[][]> mixerOutput
            , int batchId, BulletinBoardClient.ClientCallback<?> callback) {

        BatchConverter batchConverter = new BatchConverter();
        List<BulletinBoardAPI.BatchData> batchDataList = batchConverter.mixerOutput2BatchData(mixerOutput);
        asyncBulletinBoardClient.postBatch(id, batchId, batchDataList, callback);
    }

    private class BatchConverter {

        ByteString IntegerToByteString(int a){
            return ByteString.copyFrom(BigInteger.valueOf(a).toByteArray());
        }

        int ByteString2Integer(ByteString bs) {
            return Integer.valueOf(bs.toString());
        }

        List<BulletinBoardAPI.BatchData> mixerOutput2BatchData
                (Pair<Mixing.ZeroKnowledgeProof[][], Crypto.RerandomizableEncryptedMessage[][]> mixerOutput) {

            List<BulletinBoardAPI.BatchData> result = new ArrayList<BulletinBoardAPI.BatchData>();

            //

            result.add(BulletinBoardAPI.BatchData.newBuilder()
                    .setData(IntegerToByteString(n))
                    .build());

            for (Mixing.ZeroKnowledgeProof[] zkpLayer : mixerOutput.a) {
                for (Mixing.ZeroKnowledgeProof zkp : zkpLayer) {
                    result.add(BulletinBoardAPI.BatchData.newBuilder()
                            .setData(zkp.toByteString())
                            .build());
                }
            }
            for (Crypto.RerandomizableEncryptedMessage[] encryptionLayer : mixerOutput.b) {
                for (Crypto.RerandomizableEncryptedMessage encryption : encryptionLayer) {
                    result.add(BulletinBoardAPI.BatchData.newBuilder()
                            .setData(encryption.toByteString())
                            .build());
                }
            }
            return result;
        }
        Pair<Mixing.ZeroKnowledgeProof[][], Crypto.RerandomizableEncryptedMessage[][]> batchDataList2MixerOutput
                (List<BulletinBoardAPI.BatchData> batchDataList) throws Exception {

            if (n != ByteString2Integer(batchDataList.remove(0).getData())){
                throw new Exception();
            }

            int nDiv2 = n >>1;
            Mixing.ZeroKnowledgeProof[][] proofs = new Mixing.ZeroKnowledgeProof[layers][nDiv2];
            for (int layer = 0; layer <layers; layer++)
            {
                for (int proofIndex = 0 ; proofIndex < nDiv2 ; proofIndex ++)
                {
                    proofs[layer][proofIndex] = Mixing.ZeroKnowledgeProof.parseFrom(batchDataList.remove(0).getData());
                }
            }

            Crypto.RerandomizableEncryptedMessage[][] encryptions
                    = new Crypto.RerandomizableEncryptedMessage[layers + 1][n];
            for (int layer = 0; layer <layers + 1; layer++)
            {
                for (int encryptionIndex = 0 ; encryptionIndex < n ; encryptionIndex ++)
                {
                    encryptions[layer][encryptionIndex] = Crypto.RerandomizableEncryptedMessage
                            .parseFrom(batchDataList.remove(0).getData());
                }
            }

            return new Pair<Mixing.ZeroKnowledgeProof[][], Crypto.RerandomizableEncryptedMessage[][]>(proofs,encryptions);

        }

    }



    private class BatchHandler implements BulletinBoardClient.ClientCallback<SignedBatch> {

        public Pair<Mixing.ZeroKnowledgeProof[][],Crypto.RerandomizableEncryptedMessage[][]> mixerOutput;
        public boolean done;
        private int n,layers;
        private Exception e;

        public BatchHandler(int n, int layers) {
            this.n = n;
            this.layers = layers;
            done = false;
            e = null;
        }

        // convert batch message to MixerInput
        // and verify it
        @Override
        public void handleCallback(SignedBatch msg) {
            BatchConverter batchConverter = new BatchConverter();
            try {
                mixerOutput = batchConverter.batchDataList2MixerOutput(msg.getBatchDataList());
                done = verifyTable();
            } catch (Exception e) {
                this.e = e;
                done = true;
            }
        }

        @Override
        public void handleFailure(Throwable t) {

        }

        private boolean verifyTable() throws InvalidProtocolBufferException {
            int index1,index2,layer;

            //initialize locationChecksum table
            // use for check BeneshNet validity
            boolean[][] locationChecksum = new boolean[layers][n];
            for (boolean[] locationChecksumLayer: locationChecksum) {
                Arrays.fill(locationChecksumLayer,false);
            }

            Mixing.ZeroKnowledgeProof[][] zeroKnowledgeProofs = mixerOutput.a;
            Crypto.RerandomizableEncryptedMessage[][] rerandomizableEncryptedMessages = mixerOutput.b;

            for (Mixing.ZeroKnowledgeProof[] zkpLayer: zeroKnowledgeProofs) {
                for (Mixing.ZeroKnowledgeProof zkp: zkpLayer) {
                    Mixing.ZeroKnowledgeProof.Location location = zkp.getLocation();
                    index1 = location.getI();
                    index2 = location.getJ();
                    layer = location.getLayer();

                    // check location validity
                    if (layer > layers >> 1) {
                        if (index2 - index1 != n >> (layers - layer))
                            return false;
                    }
                    else{
                        if (index2 - index1 != n >> (layer + 1))
                            return false;
                    }

                    // mark location in table
                    locationChecksum[layer][index1] = true;
                    locationChecksum[layer][index2] = true;

                    // verify proof
                    if(!verifier.verify(rerandomizableEncryptedMessages[index1][layer],
                            rerandomizableEncryptedMessages[index2][layer],
                            rerandomizableEncryptedMessages[index1][layer + 1],
                            rerandomizableEncryptedMessages[index2][layer + 1],
                            zkp))
                        return  false;
                }
            }

            // verify all necessary locations for BeneshNet were proved
            for (boolean[] checksumLayer: locationChecksum) {
                for (boolean locationBoolean: checksumLayer) {
                    if (!locationBoolean)
                        return false;
                }
            }
            return true;
        }

        public List<Crypto.RerandomizableEncryptedMessage> getInputForMixer()
        {
            return Arrays.asList(mixerOutput.b[mixerOutput.b.length - 1]);
        }
    }

}
